import * as yargs from 'yargs'
import { cli } from '../../cli'
import { SSHService, SnipsServices } from '../../session'
import chalk from 'chalk'
import * as shelljs from 'shelljs'
import { HassConfiguration } from '../../models/hassConfiguration'
import { timeout } from '../../utils'

exports.command = 'skills'
exports.desc = 'Install skill present on the device'

exports.handler = async (_: yargs.Argv) => {
    let hasPythonSkills = false
    let hasGitSkills = false

    const ssh = new SSHService()
    await ssh.connect()
    .catch(e => {
        cli.stream.error(e)
        process.exit()
    })

    cli.stream.loading('Checking for skills')

    const snipsfile = await ssh.getAssistantSnipsfile()
    .catch(e => cli.stream.error('Error reading the Snipsfile for skills, reason: ' + e.message))

    const skills = await ssh.getSnippets()
    .then(snippets => {
        const skillCount = Array.from(snippets.keys()).length
        if (skillCount === 1) {
            cli.stream.success('Found 1 skill')
        } else {
            cli.stream.success(`Found ${skillCount} skills`)
        }
        return snippets
    })
    .catch(_ => cli.stream.println('No python or Home Assistant snippets found'))
    if (skills === undefined) cli.stream.println('No python or Home Assistant snippets found')

    await ssh.stopService(SnipsServices.skillServer)

    if (snipsfile !== undefined && snipsfile.skills.length > 0) {
        hasGitSkills = true
        await ssh.checkCommandExists('virtualenv')
            .then(isVirtualEnvInstalled => {
                if (!isVirtualEnvInstalled) {
                    cli.stream.loading('Installing pip package and virtualenv. This could take a while.')
                    return ssh.installPythonAndPip(_ => {})
                    .then(_ => cli.stream.success())
                }
                return
            })
            .then(_ => ssh.checkCommandExists('git'))
            .then(isGitInstalled => {
                if (!isGitInstalled) {
                    cli.stream.loading('Installing git package')
                    return ssh.aptGetY(['git'], _ => {})
                    .then(_ => cli.stream.success())
                }
                return
            })
            .then(async _ => {
                for (const skill of snipsfile.skills) {
                    if (skill.url) {
                        cli.stream.println(`Cloning skill ${skill.name} from url: ${skill.url}`)
                        await ssh.gitClone(skill.url, output => cli.stream.println(output))
                    }
                }
            })
    }

    cli.stream.println(chalk.green('Generate skills based on the snippets provided'))
    await ssh.snipsTemplateRender(output => cli.stream.print(output))

    if (skills !== undefined) {
        const skillSetup = new SkillSetup(ssh, skills)
        hasPythonSkills = skillSetup.hasPythonSkill

        if (skillSetup.hasPythonSkill) {
            cli.stream.println(chalk.green('\nSetting up Python skill'))
            await ssh.checkCommandExists('virtualenv')
            .then(isVirtualEnvInstalled => {
                if (!isVirtualEnvInstalled) {
                    cli.stream.loading('Installing pip and virtualenv. This could take a while.')
                    return ssh.installPythonAndPip(_ => {})
                    .then(_ => cli.stream.success())
                }
                return
            })
            .catch(e => cli.stream.error('Unable to install the python skill, reason: ' + e.message))
        }

        if (skillSetup.hasHassSkill) {
            cli.stream.println(chalk.green('\nSetting up Home Assistant skill'))
            await skillSetup.setupHomeAssistantSkill()
            .then(_ => cli.stream.loading('Relaunching Home Assistant'))
            .then(_ => ssh.restartHomeAssistant())
            .then(_ => cli.stream.success())
            .catch(e => cli.stream.error('Unable to install Home Assistant skill, reason: ' + e.message))
        }
    }

    if (hasPythonSkills || hasGitSkills) {
        cli.stream.loading('Generating virtual environment for python skills')

        await ssh.runSkillSetupSH()
        .then(_ => cli.stream.success())
        .then(_ => cli.stream.println('Relaunching snips-skill-server'))
        .then(_ => ssh.relaunchService(SnipsServices.skillServer))
        .catch(e => cli.stream.error('Error setting up virtualenv, reason : ' + e.message))
    }

    const assistant = await ssh.getInstalledAssistant()
    if (assistant !== undefined) {
        cli.stream.success(`Snips assistant is now running. Say ${chalk.blue(assistant.hotword)} to start!`)
    } else {
        cli.stream.success(`Snips assistant is now running`)
    }

    ssh.disconnect()
    process.exit()
}

class SkillSetup {

    ssh: SSHService

    hasPythonSkill: boolean
    hasHassSkill: boolean

    hassSnippets: Map<string, string[]>

    hassCfgPath = '/home/homeassistant/.homeassistant/'

    constructor(ssh: any, snippets: Map<string, { type: string, snippets: string[] }>) {
        this.ssh = ssh
        this.hassSnippets = new Map<string, string[]>()

        snippets.forEach((skill, key) => {
            if (skill.type.includes('python')) this.hasPythonSkill = true
            if (skill.type.includes('homeassistant')) {
                this.hasHassSkill = true
                this.hassSnippets.set(key, skill.snippets)
            }
        })
    }

    async setupHomeAssistantSkill() {
        // Check for existing Hass configuration file
        await this.ssh.fileExists(this.hassCfgPath + 'configuration.yaml')
        .then(exists => {
            if (!exists) return this.askUserForCustomHassInstallConfigurationPath()
            cli.stream.success('Found Home Assistant configuration.yaml')
            return this.hassCfgPath
        })
        .then(cfgFilePath => {
            if (cfgFilePath === undefined) return this.installHass()
            return cfgFilePath
        })
        .then(cfgFilePath => this.hassCfgPath = cfgFilePath.replace(/\/?$/, '/'))
        .then(_ => cli.stream.println('Moving generated skills to Hass python_script folder'))
        .then(_ => this.ssh.moveSkillsToHomeAssistantFolder(this.hassCfgPath + 'python_scripts', Array.from(this.hassSnippets.keys())))
        .then(_ => cli.stream.println('Backing up your last Hass configuration'))
        .then(_ => this.ssh.backupHassConfiguration(this.hassCfgPath))
        .then(_ => this.ssh.getFile(this.hassCfgPath + 'configuration.yaml'))
        .then(file => {
            const parsedYamlFile = HassConfiguration.parse(file)
            if (parsedYamlFile === undefined) throw new Error('Error parsing the Home Assistant configuration')
            return parsedYamlFile
        })
        .then(hassCfgFile => {
            cli.stream.println('Creating new Hass configuration')
            return this.ssh.getInstalledAssistant()
            .then(assistantFile => {
                return HassConfiguration.generateSnipsConfiguration(hassCfgFile, assistantFile, this.hassSnippets)
            })
        })
        .then(newYamlFilePath => {
            return this.ssh.putFile(newYamlFilePath, '/tmp/configuration.yaml')
            .then(_ => shelljs.exec(`rm ${newYamlFilePath}`, { async: false, silent: false }))
            .then(_ => this.ssh.run(`sudo mv /tmp/configuration.yaml ${this.hassCfgPath}`))
        })
    }

    private async askUserForCustomHassInstallConfigurationPath(): Promise<string | undefined> {
        return cli.prompt.promptQuestions([
            {
                name: 'isInstalled',
                type: 'confirm',
                message: 'Is Home Assistant already installed on this device?',
            },
        ])
        .then(isHassInstalled => {
            if (isHassInstalled['isInstalled']) {
                let hassCfgFound = false
                return cli.prompt.promptQuestions([
                    {
                        name: 'hassCfgPath',
                        type: 'input',
                        message: 'Enter the path to Home Assistant configuration file.',
                        validate: async value => {
                            if (value.length === 0) return `Path can't be blank`
                            const pathWithSlash = value.replace(/\/?$/, '/')
                            hassCfgFound = await this.ssh.fileExists(pathWithSlash + 'configuration.yaml')
                            if (!hassCfgFound) return `Didn't find ${pathWithSlash}configuration.yaml on the device`
                            return true
                        },
                    },
                ])
                .then(cfgPathResponse => cfgPathResponse['hassCfgPath'])
            }
            return undefined
        })
    }

    private async installHass(): Promise<string> {
        const hassInstallation = await cli.prompt.promptQuestions([
            {
                name: 'installHass',
                type: 'confirm',
                message: 'You need to install Home Assistant to continue, proceed?',
            },
        ])
        if (!hassInstallation['installHass']) {
            cli.stream.println(`Quitting`)
            cli.stream.hint('You can install Home Assistant by running ' + chalk.blue('sam hass install'))
            throw new Error(`Hass isn't installed, aborting`)
        }

        cli.stream.println('Installing Home Assistant, this can take a while')
        return await this.ssh.installHomeAssistant(
            _ => {},
            stepTitle => {
                cli.stream.println(chalk.green(stepTitle))
            })
            .then(async _ => {
                let loadingText = 'Waiting for Home Assistant initial setup. This could take a while.'
                cli.stream.loading(loadingText)
                let homeAssistantConfigurationDetected = false
                while (homeAssistantConfigurationDetected === false) {
                    await timeout(5000)
                    homeAssistantConfigurationDetected = await this.ssh.fileExists('/home/homeassistant/.homeassistant/configuration.yaml')
                    loadingText += ' . '
                    cli.stream.loading(loadingText)
                }
                cli.stream.success()
                return 'success'
            })
            .then(_ => this.hassCfgPath)
    }
}

