import chalk from 'chalk'
import * as yargs from 'yargs'
import { cli } from '../../cli'
import { AudioDevice } from '../../utils/audiodevice'
import { SSHService, SnipsServices } from '../../session/ssh'

exports.command = 'audio'
exports.desc = 'Quickly setup the mic & speaker on your device'

exports.handler = async (_: yargs.Argv) => {
    const ssh = new SSHService()
    await ssh.connect()
    .catch(e => {
        cli.stream.error(e)
        process.exit()
    })

    let captureDevice: AudioDevice | undefined
    try {
        const captureDevices = await ssh.listCaptureDevices()
        if (captureDevices.length === 1) {
            cli.stream.success(`Found only 1 capture device named: ${captureDevices[0].name}`)
            cli.stream.println(`We will use this device to capture sound`)
            captureDevice = captureDevices[0]
        } else {
            const response = await cli.prompt.promptList(
                'device',
                `We found ${captureDevices.length} interfaces to record sound.\n  Choose the one you want to use`,
                captureDevices.map(device => device.name),
            )
            captureDevice = captureDevices.find(device => device.name === response['device'])
        }
    } catch (e) {
        cli.stream.error('Error fetching audio capture device: ' + e.message)
        cli.stream.hint('If the microphone is not connected through USB, you might need to install it yourself. See the manufacturers installation guide')
    }
    if (captureDevice === undefined) {
        cli.stream.hint('Using default capture configuration card 1 subdevice 0')
        captureDevice = { name: 'Default', card: 1, subdevice: 0 }
    }

    let outputDevice: AudioDevice | undefined
    try {
        const outputDevices = await ssh.listOuputDevices()
        if (outputDevices.length === 1) {
            cli.stream.success(`Found only 1 output device named: ${outputDevices[0].name}`)
            cli.stream.println(`We will use this device to output sound`)
            outputDevice = outputDevices[0]
        } else {
            const response = await cli.prompt.promptList(
                'device',
                `We found ${outputDevices.length} interfaces to output sound.\n  Choose the one you want to use as your speaker`,
                outputDevices.map(device => device.name),
            )
            outputDevice = outputDevices.find(device => device.name === response['device'])
        }
    } catch (e) {
        cli.stream.error('Error fetching audio output device: ' + e.message)
    }
    if (outputDevice === undefined) {
        cli.stream.error(`An output device wasn't found`)
        return
    }

    try {
        await ssh.setupAsoundConf(outputDevice, captureDevice)
        cli.stream.success('Installed /etc/asound.conf')
    } catch (e) {
        cli.stream.error('Error installing the microphone asound.conf: ' + e.message)
        ssh.disconnect()
        process.exit()
    }

    cli.stream.println('Setting to volume to 90%')
    await ssh.setVolumeLevel(outputDevice.card, 90)
    .catch(e => cli.stream.error('Error setting up the volume: ' + e.message))

    cli.stream.loading('Restarting snips-audio-server service')
    await ssh.relaunchService(SnipsServices.audioServer).catch(_ => {})
    cli.stream.done()

    cli.stream.hint(chalk`You can test your microphone & speaker with {blue sam test microphone / speaker}`)
    ssh.disconnect()
}
