import fsExtra = require('fs-extra')
import { DeviceCredentials, AssistantFile, Snipsfile } from '../models'
import { AudioDevice, newError, timeout } from '../utils'
import { Assets } from '../assets'
import { cli } from '../cli'
import * as nodeSsh from 'node-ssh'
import * as shelljs from 'shelljs'
import os = require('os')
import path = require('path')

export const snipsRemoteAuthorizedKeysFilename = '~/.ssh/authorized_keys'

const snipsRSAKeyName = 'id_rsa_snips'
const snipsRSAKeyFilePath = path.join(os.homedir(), '.ssh', snipsRSAKeyName)

export enum SnipsServices {
    analytics = 'snips-analytics',
    asr = 'snips-asr',
    audioServer = 'snips-audio-server',
    dialogue = 'snips-dialogue',
    hotword = 'snips-hotword',
    nlu = 'snips-nlu',
    skillServer = 'snips-skill-server',
    template = 'snips-template',
    tts = 'snips-tts',
    watch = 'snips-watch',
}

export const availableSnipsServices: string[] = [
    SnipsServices.analytics,
    SnipsServices.asr,
    SnipsServices.audioServer,
    SnipsServices.dialogue,
    SnipsServices.hotword,
    SnipsServices.nlu,
    SnipsServices.skillServer,
    SnipsServices.template,
    SnipsServices.tts,
    SnipsServices.watch,
]

export class SSHService {

    ssh: any
    credentials: DeviceCredentials

    constructor() {
        this.ssh = new nodeSsh()
    }

    public async connect(): Promise<string> {
        const credentials: DeviceCredentials | undefined = cli.config.credentials
        if (credentials === undefined ||
            credentials.hostname === null) {
            throw new Error(`No credentials found`)
        }

        return this.connectInternal(credentials)
        .catch(_ => {
            this.disconnect()
            throw newError('Connection failed, is your device turned on?')
        })
    }

    public async retryConnect(tick: () => void, limit = 5): Promise<boolean> {
        let isConnected = false
        let numberOfRetries = 0
        while (isConnected === false && limit > numberOfRetries) {
            await timeout(5000)
            tick()
            await this.connect()
            .then(_ =>  isConnected = true)
            .catch(_ => {})
            numberOfRetries += 1
        }
        return isConnected
    }

    public async newConnection(credentials: DeviceCredentials, password: string, onConnected: () => void, onSSHCopyId: () => void, onCredentialsSaved: () => void): Promise<void> {
        await this.connectInternal(credentials, password)
        .then(_ => onConnected())
        .catch(e => {
            throw new Error(`Error connecting to ${credentials.hostname}: ${e.message}`)
        })

        await this.sshCopyId()
        .then(_ => onSSHCopyId())
        .catch(e => {
            this.disconnect()
            throw new Error(`Failed to copy SSH public key: ${e.message}`)
        })

        cli.config.credentials = this.credentials
        await cli.config.save()
        .then(_ => onCredentialsSaved())
        .catch(e => {
            this.disconnect()
            throw new Error(`Failed to save credentials: ${e.message}`)
        })
    }

    private async connectInternal (credentials: DeviceCredentials, password?: string): Promise<any> {
        this.credentials = credentials

        if (password != null) {
            return this.ssh.connect({
                password,
                host: this.credentials.hostname,
                username: this.credentials.username,

            })
        }
        return this.ssh.connect({
            host: this.credentials.hostname,
            username: this.credentials.username,
            privateKey: snipsRSAKeyFilePath,
        })
    }

    public async disconnect(): Promise<void> {
        await this.ssh.dispose()
    }

    public async reboot(): Promise<void> {
        this.ssh.exec('sudo reboot')
    }

    public async isAssistantInstalled(): Promise<boolean> {
        return this.ssh.exec(`if [ -d /usr/share/snips/assistant ]; then echo 1; else echo 0; fi`)
        .then(result => result === '1')
    }

    public async getInstalledAssistant(): Promise<AssistantFile> {
        const assistantFilename = '/usr/share/snips/assistant/assistant.json'
        return this.ssh.exec(`if [ -f ${assistantFilename} ]; then cat ${assistantFilename}; else echo ""; fi`)
        .then(assistantJSON => {
            if (assistantJSON === '') throw new Error('No assistant installed')
            return new AssistantFile(assistantJSON)
        })
    }

    public async getOSVersion(): Promise<string> {
        return this.ssh.exec('cat /etc/os-release')
        .then(success => {
            const osVersion = success.split('\n')[0].replace('PRETTY_NAME="', '').replace('"', '')
            if (osVersion === null || osVersion.length === 0) {
                return 'Unknown'
            }
            return osVersion
        })
        .catch(_ => {
            return 'Unknown'
        })
    }

    public async installSnipsPlatformDemo(listener: (output: string) => void): Promise<string> {
        return this.aptGetY(['snips-platform-demo'], listener)
    }

    public async aptGetY(packageNames: string[], listener: (output: string) => void): Promise<string> {
        return this.sshCommandLog(`sudo apt-get install -y ${packageNames.join(' ')}`, listener)
    }

    public async isNetcatInstalled(): Promise<boolean> {
        return this.ssh.exec(`if [ -x "$(command -v nc)" ]; then echo 1; else echo 0; fi`)
        .then(result => result === '1')
        .then(result => {
            if (!result) {
                return this.dpkgCheckIfPackageIsInstalled('netcat')
            }
            return result
        })
    }

    public async speakerTest(): Promise<string> {
        return this.ssh.exec(`sudo speaker-test -c2 -twav`)
    }

    public async stopSpeakerTest(): Promise<string> {
        return this.ssh.exec(`pkill speaker-test`)
    }

    public async setHostname(newHostname: string): Promise<string> {
        const oldHostname = await this.ssh.exec('cat /etc/hostname')
        .catch(_ => {
            throw new Error('You device hostname was not readable')
        })

        return this.ssh.exec(`sudo sed -i "s/${oldHostname}/${newHostname}/g" /etc/hosts; sudo sed -i "s/${oldHostname}/${newHostname}/g" /etc/hostname; sudo reboot`)
        .then(_ => {
            this.credentials.hostname = newHostname + '.local'
            cli.config.credentials = this.credentials
            return cli.config.save()
        })
        .catch(_ => {
            this.credentials.hostname = newHostname + '.local'
            cli.config.credentials = this.credentials
            return cli.config.save()
        })
    }

    public static async fetchBootHostname(ips: string[]): Promise<{ ip: string, hostname: string, snipsHostname: string }[]> {
        return await Promise.all(
            ips.map(async ip => {
                const client = new nodeSsh()
                const hostnames = await client.connect({
                    host: ip,
                    username: 'pi',
                    password: 'raspberry',
                })
                .then(async _ => {
                    const snipsHostname = await client.exec(`if [ -f /boot/alias ]; then echo 1; else echo 0; fi`)
                    .then(result => {
                        if (result === '1') {
                            return client.exec('cat /boot/alias').catch(_ => '')
                        }
                        return ''
                    })
                    const hostname: string = await client.exec('cat /etc/hostname').catch(_ => '')
                    return { hostname, snipsHostname }
                })
                .catch(_ => {
                    return { snipsHostname: '', hostname: '' }
                })
                return { ip, hostname: hostnames.hostname, snipsHostname: hostnames.snipsHostname }
            }),
        )
    }

    public async getCaptureLevel(card: number | undefined = 1): Promise<string> {
        return this.ssh.exec(`amixer -c ${card} cget name='Mic Capture Volume'`)
    }

    public async setCaptureLevel(card: number | undefined = 1, level: number | undefined = 80): Promise<string> {
        return this.ssh.exec(`amixer -c ${card} cset name='Mic Capture Volume' ${level}%`)
    }

    public async getVolumeLevel(card: number | undefined = 0): Promise<string> {
        return this.ssh.exec(`amixer -c ${card} cget name='PCM Playback Volume'`)
    }

    public async setVolumeLevel(card: number | undefined = 0, level: number | undefined = 90): Promise<string> {
        return this.ssh.exec(`amixer -c ${card} cset name='PCM Playback Volume' ${level}%`)
    }

    public async recordAudio(name: string): Promise<string> {
        return this.ssh.exec(`arecord -f cd ${name}.wav`)
    }

    public async stopRecording(): Promise<string> {
        return this.ssh.exec(`pkill arecord`)
    }

    public async testingtesting(): Promise<string> {
        return this.ssh.exec('sudo apt-get install ')
    }

    public async listCaptureDevices(): Promise<AudioDevice[]> {
        return this.ssh.exec(`arecord -l | grep -E '^[^  ][^**]'`)
        .then(success => {
            // card 0: ALSA [bcm2835 ALSA], device 0: bcm2835 ALSA [bcm2835 ALSA]
            if (success.length === 0) {
                throw new Error('No capture devices found. Check if your mic is correctly plugged.')
            }
            const devices = String(success).split('\n')
            if (devices.length === 0) {
                throw new Error('No capture devices found. Check if your mic is correctly plugged.')
            }
            try {
                return devices.map(device => new AudioDevice(device))
            } catch (e) {
                throw new Error('Error reading the information of the audio device: ' + e.message)
            }
        })
    }

    public async listOuputDevices(): Promise<AudioDevice[]> {
        return this.ssh.exec(`aplay -l | grep -E '^[^  ][^**]'`)
        .then(success => {
            const devices = String(success).split('\n')
            if (devices.length === 0) {
                throw new Error('No output devices found. This should not happen, on your device, run \narecord -l')
            }
            try {
                return devices.map(device => new AudioDevice(device))
            } catch (e) {
                throw new Error('Error reading the information of the audio device: ' + e.message)
            }
        })
    }

    public async launchSnipsWatch(listener: (string) => void): Promise<string> {
        return this.sshCommandLog(`${SnipsServices.watch} -vvv`, listener)
    }

    public async journalctlLogs(services: string[] | undefined, listener: (string) => void): Promise<string> {
        if (services !== undefined && services.length > 0) {
            const s = services.map(service => `-u ${service}`).join(' ')
            return this.sshCommandLog(`journalctl -f ${s}`, listener)
        }
        return this.sshCommandLog(`journalctl -f -u "snips-*"`, listener)
    }

    public async snipsServicesSystemctlStatus(): Promise<{ name: string, active: boolean }[]> {
        return Promise.all(availableSnipsServices.map(async serviceName => {
            const active = await this.systemctlStatus(serviceName)
            return { active, name: serviceName }
        }))
    }

    public async systemctlStatus(service: string): Promise<boolean> {
        return await this.ssh.exec(`systemctl is-active ${service}.service >/dev/null 2>&1 && echo 'active' || echo 'inactive'`)
            .then(result => result === 'active')
    }

    public async playAudio(name: string): Promise<string> {
        // Playing audio throws back an error even though it executes the file. We ignore the error.
        return this.ssh.exec(`aplay ${name}.wav`)
        // .catch((e) => { });
    }

    public async removeAudioFile(name: string): Promise<string> {
        return this.ssh.exec(`rm ${name}.wav`)
    }

    public async setupAsoundConf(outputDevice: AudioDevice, captureDevice: AudioDevice): Promise<string> {
        return this.ssh.exec(`(echo 'pcm.!default {'; \
        echo '  type asym'; \
        echo '  playback.pcm {'; \
        echo '    type plug'; \
        echo '    slave.pcm "hw:${outputDevice.card},${outputDevice.subdevice}"'; \
        echo '  }'; \
        echo '  capture.pcm {'; \
        echo '    type plug'; \
        echo '    slave.pcm "hw:${captureDevice.card},${captureDevice.subdevice}"'; \
        echo '  }'; \
        echo '}') \
        | sudo tee /etc/asound.conf > /dev/null`)
    }

    public async getAssistantSnipsfile(): Promise<Snipsfile> {
        return this.getFile('/usr/share/snips/Snipsfile.yaml')
        .then(snipsfile => new Snipsfile(snipsfile))
    }

    public async copySnipsfile(snipsfile: string) {
        return this.ssh.exec(`echo "${snipsfile}" | sudo tee ~/Snipsfile > /dev/null`)
    }

    public async stopService(service: string): Promise<string> {
        return this.ssh.exec(`sudo systemctl stop ${service}.service`)
    }

    public async relaunchService(service: string): Promise<string> {
        return this.ssh.exec(`sudo systemctl restart ${service}.service`)
    }

    public async relaunchServices(services: string[] | undefined): Promise<string> {
        if (services !== undefined && services.length > 0) {
            const restart = services.map(service => `sudo systemctl restart ${service}`).join(' ; ')
            return this.ssh.exec(restart)
        }
        return this.ssh.exec(`sudo systemctl restart 'snips*'`)
    }

    public async updateSnips(listener: (string) => void): Promise<string> {
        return this.sshCommandLog(`sudo apt-get update; sudo apt-get -o Dpkg::Options::="--force-confold" upgrade -y snips-platform-voice ${SnipsServices.template} ${SnipsServices.skillServer}`, listener)
    }

    public async installSnips(listener: (string) => void, installSteps: (string) => void): Promise<string> {
        installSteps('Updating aptitude repository')
        return this.sshCommandLog('sudo apt-get update', listener)
        .then(_ => {
            installSteps('Installing dirmngr')
            return this.sshCommandLog('sudo apt-get install -y dirmngr', listener)
            .catch(_ => { throw new Error(`Couldn't install dirmngr`) })
        })
        .then(_ => {
            installSteps('Adding Snips APT source list')
            return this.sshCommandLog(`sudo bash -c  'echo "deb https://raspbian.snips.ai/$(lsb_release -cs) stable main" > /etc/apt/sources.list.d/snips.list'`, listener)
            .catch(_ => { throw new Error(`Couldn't update Snips apt sources on the device`) })
        })
        .then(_ => {
            installSteps('Adding Snips PGP public key')
            return this.ssh.exec(`(echo "${Assets.pgpKey()}" > /tmp/snips-pgp-key.asc) && sudo apt-key add /tmp/snips-pgp-key.asc`)
            .catch(_ => {
                // PGP will complain about the PGP key already being added or that the apt-key output should not be parsed
                // We can't safely catch these errors since it's language system dependant
            })
        })
        .then(_ => installSteps('Updating aptitude repository'))
        .then(_ => this.sshCommandLog('sudo apt-get update', listener))
        .then(_ => installSteps('Installing Snips Platform, this could take a while, grab a cup of tea & relax'))
        .then(_ => {
            return this.sshCommandLog(`sudo apt-get install -y snips-platform-voice ${SnipsServices.watch} ${SnipsServices.template} ${SnipsServices.skillServer}`, listener)
            .catch(err => { throw new Error(`There was a problem during Snips Platform installation ${err}`) })
        })
        .then(_ => installSteps('Adding current user to snips-skills-admin group for security'))
        .then(_ => this.addCurrentUserToSnipsSkillsAdminGroup())
    }

    public async packagesVersion(): Promise<{ name: string, version: string }[]> {
        const formatOutput = (dpkgOutput: string): { name: string, version: string }[] => {
            return dpkgOutput
            .split('\n')
            .map(line => {
                const split = line.trim().split(' ')
                if (split.length >= 2) return { name: split[0], version: split[split.length - 1] }
                return { name: '', version: '' }
            })
            .filter(v => v.name !== '')
        }
        return this.ssh.exec(`dpkg -l | grep snips | awk '{printf("%- 30s %- 10s\\n", $2,$3)}'`)
        .then(output => formatOutput(output).filter(v => availableSnipsServices.includes(v.name)))
    }

    public async updateWifi(name: string, passphrase: string): Promise<string> {
        return this.ssh.exec(`(echo ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev; \
            echo update_config=1; \
            echo ; \
            wpa_passphrase \"${name}\" \"${passphrase}\") \
            | sudo tee /etc/wpa_supplicant/wpa_supplicant.conf > /dev/null`)
            .then(_ => {
                return this.ssh.exec('sudo wpa_cli reconfigure')
            })
    }

    public async statusWifi(): Promise<string> {
        return this.ssh.exec('sudo wpa_cli status')
        .then(response => {
            let ssid: string | undefined
            let status: string | undefined
            response.split('\n')
            .map(line => line.split('='))
            .forEach(lineComponents => {
                const key = lineComponents[0]
                const value = lineComponents[1]
                if (key === 'ssid') {
                    ssid = value
                } else if (key === 'wpa_state') {
                    // TODO add wpa_state
                    status = value
                }
            })
            return [ssid, status]
        })
    }

    public async reconnectWifi(): Promise<string> {
        return this.ssh.exec('sudo wpa_cli reconnect')
    }

    public async disconnectWifi(): Promise<string> {
        return this.ssh.exec('sudo wpa_cli disconnect')
    }

    public async putFile(localPath: string, remotePath: string): Promise<string> {
        return this.ssh.putFile(localPath, remotePath)
    }

    public async putString(string: string, remotePath: string) {
        return this.ssh.exec(`sudo sh -c "echo '${string}' > ${remotePath}"`)
    }

    public async getFile(remotePath: string): Promise<string> {
        return this.ssh.exec(`cat ${remotePath}`)
    }

    public async putDirectory(localDirectory: string, remoteDirectory: string): Promise<string> {
        return await this.ssh.putDirectory(localDirectory, remoteDirectory, {
            recursive: true,
            concurrency: 1, // /!\ Anything >1 will result in potentially failed transfered files, you've been warned
            validate: itemPath => {
                const baseName = path.basename(itemPath)
                return baseName.substr(0, 1) !== '.' // do not allow dot files
            },
        })
    }

    public async removeFile(remotePath: string): Promise<string> {
        return this.ssh.exec(`sudo rm ${remotePath}`)
    }

    public async removeDirectory(remotePath: string): Promise<string> {
        return this.ssh.exec(`sudo rm -rf ${remotePath}`)
    }

    public async unzipFile(filePath: string, destinationPath: string = '.'): Promise<string> {
        // Unzip overwrites already present files
        return this.ssh.exec(`unzip -o ${filePath} -d ${destinationPath}`)
    }

    public async installAssistant(assistantPath: string): Promise<string> {
        const assistantFilename: string = path.basename(assistantPath)
        const tmpDirPath = '/tmp/assistant/'

        return this.removeDirectory('/usr/share/snips/assistant')
        .then(_ => {
            if (assistantFilename.includes('zip')) {
                return this.removeDirectory(tmpDirPath)
                .then(_ => this.ssh.exec(`mkdir ${tmpDirPath}`))
                .then(_ => this.putFile(assistantPath, tmpDirPath + assistantFilename))
                .then(_ => this.unzipFile(tmpDirPath + assistantFilename, tmpDirPath))
                .then(_ => this.removeFile(tmpDirPath + assistantFilename))
            }
            return this.putDirectory(assistantPath, tmpDirPath)
            .then(status => {
                // PutDirectory is not sending an explicit error, so we catch it here.
                if (!status) {
                    throw new Error('Some files failed to transfer, aborting')
                }
                return 'success'
            })
        })
        .then(_ => this.ssh.exec(`sudo mv ${tmpDirPath}* /usr/share/snips`))
    }

    public async dpkgCheckIfPackageIsInstalled(packageName: string): Promise<boolean> {
        return this.ssh.exec(`if dpkg-query -W -f='\${db:Status-abbrev}' "${packageName}" 2>/dev/null | grep -q '^i'; then echo 1; else echo 0; fi`)
        .then(result => result === '1')
    }

    public async isFileExecutable(file: string): Promise<boolean> {
        return this.ssh.exec(`if [ -x "${file}" ]; then echo 1; else echo 0; fi`)
        .then(result => result === '1')
    }

    public async checkCommandExists(command: string): Promise<boolean> {
        return this.ssh.exec(`if [ -x "$(command -v ${command})" ]; then echo 1; else echo 0; fi`)
        .then(result => result === '1')
    }

    public async installRespeakerMicHat(listener: (string) => void): Promise<string> {
        return this.sshCommandLog('git clone https://github.com/respeaker/seeed-voicecard.git', listener)
        .then(_ => this.sshCommandLog('cd seeed-voicecard; sudo ./install.sh 2mic; sudo reboot', listener))
    }

    public async run(command: string): Promise<string> {
        return this.ssh.exec(command)
    }

    private async sshCommandLog(command: string, listener: (string) => void): Promise<string> {
        return this.ssh.execCommand(command, {
            onStdout(chunk) {
                listener(chunk.toString('utf8'))
            },
            onStderr(chunk) {
                listener(chunk.toString('utf8'))
            },
        })
    }

    /*----------------------------------------------------------------------*/
    // Skill store related

    public async snipsTemplateRender(listener: (string) => void): Promise<string> {
        return this.sshCommandLog(`${SnipsServices.template} render`, listener)
    }

    public async addCurrentUserToSnipsSkillsAdminGroup(): Promise<string> {
        return this.ssh.exec(`if id -nG "$USER" | grep -qw "snips-skills-admin"; then echo 1; else echo 0; fi`)
        .then(result => result === '1')
        .then(isInGroup => {
            if (!isInGroup) return this.ssh.exec('sudo usermod -a -G snips-skills-admin $USER')
            return 'done'
        })
    }

    public async gitClone(repository: string, listener: (string) => void) {
        const gitFolderName = path.basename(repository).replace(/\.[^/.]+$/, '')
        return this.sshCommandLog(`cd /var/lib/snips/skills/; if [ ! -d "${gitFolderName}" ]; then git clone --recursive "${repository}" "${gitFolderName}"; else cd "${gitFolderName}"; git pull; fi;`, listener)
        .then(_ => this.ssh.exec(`find /var/lib/snips/skills/${gitFolderName} -name action* ! -executable`))
        .then((actionsNotExecutable: string) => {
            if (actionsNotExecutable.length > 0) {
                listener('Warning: found action files that are not executable. Changing their permissions')
                listener(actionsNotExecutable + '\n')
                const actions = actionsNotExecutable.split('\n').join(' ')
                return this.ssh.exec(`chmod +x ${actions}`)
            }
        })
    }

    public async runSkillSetupSH(): Promise<string> {
        return this.ssh.exec('find /var/lib/snips/skills -type f -name setup.sh -execdir sh setup.sh {} \\;')
    }

    public async getSnippets(): Promise<Map<string, { type: string, snippets: string[] }>> {
        return this.ssh.exec('cd /usr/share/snips/assistant/snippets; ls -d */*/*')
        .then((snippets: string) => {
            // returns for instance :
            // Smart_lights/python2/lightsTurnDown.snippet
            // Smart_lights/python2/lightsTurnOff.snippet
            if (snippets === undefined) throw new Error('No snippets found')
            if (snippets.length === 0) throw new Error('No snippets found')

            return snippets.split('\n')
            .map(line => {
                const parts = line.split('/')
                if (parts.length === 0 || parts.length < 3) throw new Error('No snippets found')

                return {
                    name: parts[0],
                    type: parts[1],
                    snippet: parts[2].replace(/.snippet/g, ''),
                }
            })
            .reduce((previous, next) => {
                let subArray = previous.get(next.name)
                if (subArray !== undefined) {
                    subArray.snippets.push(next.snippet)
                } else {
                    subArray = { type: next.type, snippets: [next.snippet] }
                }
                previous.set(next.name, subArray)
                return previous
            },      new Map<string, { type: string, snippets: string[] }>())
        })
    }

    public async getGeneratedSkills(): Promise<Map<string, string[]>> {
        // Find all action files excluding the venv files
        // The ssh query returns a list of path:
        // ./Smart_lights/lightsTurnOn
        // ./Smart_lights/lightsTurnOff
        return this.ssh.exec(`cd /var/lib/snips/skills; find . -name venv* -prune -o -name 'action*' -print`)
        .then((skillPath: string) => {
            return skillPath.split('\n')
            .map(line => {
                return {
                    name: path.dirname(line).replace(/.\//g, ''),
                    action: path.basename(line),
                }
            })
            .reduce((previous, next) => {
                const subArray = previous.get(next.name) ? previous.get(next.name) : []
                subArray!.push(next.action)
                previous.set(next.name, subArray!)
                return previous
            },      new Map<string, string[]>())
        })
    }

    /*----------------------------------------------------------------------*/
    // Home assistant related

    public async moveSkillsToHomeAssistantFolder(pythonFolder: string, skillNames: string[]) {
        return this.folderExists(pythonFolder)
        .then(pythonFolderExists => {
            if (!pythonFolderExists) return this.ssh.exec(`sudo mkdir ${pythonFolder}`)
            return ''
        })
        .then(async _ => {
            const moveFolders = skillNames.map(name => `/var/lib/snips/skills/${name}/*.py`).join(' ')
            return this.ssh.exec(`sudo mv -t ${pythonFolder} ${moveFolders}`)
        })
    }

    public async readHASSConfig(): Promise<string> {
        return this.ssh.exec('')
    }

    public async backupHassConfiguration(hassConfigurationPath: string): Promise<string> {
        return this.ssh.exec(`cd ${hassConfigurationPath}; sudo cp configuration.yaml configuration.yaml.bak-$(date +%-Y%-m%-d)-$(date +%-T)`)
    }

    public async folderExists(remoteFilePath: string): Promise<boolean> {
        return this.ssh.exec(`if [ -d "${remoteFilePath}" ]; then echo 1; else echo 0; fi`)
        .then(result => result === '1')
    }

    public async fileExists(remoteFilePath: string): Promise<boolean> {
        return this.ssh.exec(`if [ -f "${remoteFilePath}" ]; then echo 1; else echo 0; fi`)
        .then(result => result === '1')
    }

    public async restartHomeAssistant(): Promise<string> {
        return this.ssh.exec('sudo systemctl restart home-assistant@homeassistant.service')
    }

    public async updateHomeAssistant(listener: (string) => void) {
        return this.sshCommandLog(this.hassBashCmd('cd /srv/homeassistant; source bin/activate; pip3 install --upgrade homeassistant'), listener)
    }

    public async installPythonAndPip(listener: (string) => void) {
        return this.sshCommandLog('sudo apt-get install -y python3-pip python3-venv python-pip; sudo pip install virtualenv', listener)
    }

    public async installHomeAssistant(listener: (string) => void, installStep: (string) => void): Promise<string> {
        installStep('Updating APT repositories')
        return this.sshCommandLog('sudo apt-get update', listener)
        .then(_ => installStep('Installing python3-pip, python3-venv'))
        .then(_ => this.sshCommandLog('sudo apt-get install -y python3-pip python3-venv', listener))
        .then(_ => installStep('Adding homeassistant user'))
        .then(_ => this.sshCommandLog('sudo useradd -rm homeassistant', listener))
        .then(_ => this.sshCommandLog('cd /srv; sudo mkdir homeassistant; sudo chown homeassistant:homeassistant homeassistant', listener))
        .then(_ => installStep('Installing Home Assistant'))
        .then(_ => this.sshCommandLog(this.hassBashCmd('cd /srv/homeassistant; python3 -m venv .; source bin/activate; python3 -m pip install wheel; pip3 install homeassistant'), listener))
        .then(_ => installStep('Adding Hass systemd service to enable autostart'))
        .then(_ => this.sshCommandLog(`echo '${Assets.homeAssistantPythonSystemdService()}' | sudo tee /etc/systemd/system/home-assistant@homeassistant.service > /dev/null`, listener))
        .then(_ => this.ssh.exec('sudo systemctl --system daemon-reload; sudo systemctl enable home-assistant@homeassistant.service')
                    .catch(_ => {}))
        .then(_ => installStep(`Launching Hass, it can take a while the first time. Check http://${this.credentials.hostname}:8123 to see when it's up.`))
        .then(_ => this.ssh.exec('sudo systemctl start home-assistant@homeassistant.service'))
    }

    private hassBashCmd(command: string): string {
        return `sudo su -c '${command}' -s /bin/bash homeassistant`
    }

    /*----------------------------------------------------------------------*/
    // ASR Benchmark

    public async playWAVtoASR(remoteWAVPath: string): Promise<string> {
        return this.ssh.exec(`cat ${remoteWAVPath} | nc localhost 1234`)
    }

    public async journalctlLog(service: string, listener: (string) => void): Promise<string> {
        return this.sshCommandLog(`journalctl -S now -o cat -f -u ${service}.service`, listener)
    }

    public async editAudioServerService(): Promise<string> {
        return this.ssh.exec(`sudo sed -i "s#.*ExecStart=/usr/bin/${SnipsServices.audioServer}.*#ExecStart=/usr/bin/${SnipsServices.audioServer} --hijack localhost:1234 --nomike#g" /lib/systemd/system/${SnipsServices.audioServer}.service; \
        sudo systemctl daemon-reload; \
        sudo systemctl restart ${SnipsServices.audioServer}.service`)
    }

    public async resetAudioServerService(): Promise<string> {
        return this.ssh.exec(`sudo sed -i "s#.*ExecStart=/usr/bin/${SnipsServices.audioServer}.*#ExecStart=/usr/bin/${SnipsServices.audioServer}#g" /lib/systemd/system/${SnipsServices.audioServer}.service; \
        sudo systemctl daemon-reload; \
        sudo systemctl restart ${SnipsServices.audioServer}.service`)
    }

    public async editASRService(maxActive: number, minActive: number): Promise<string> {
        return this.ssh.exec(`sudo sed -i "s#.*ExecStart=/usr/bin/${SnipsServices.asr}.*#ExecStart=/usr/bin/${SnipsServices.asr} --max_active ${maxActive} --min_active ${minActive}#g" /lib/systemd/system/${SnipsServices.asr}.service; \
        sudo systemctl daemon-reload; \
        sudo systemctl restart ${SnipsServices.asr}.service`)
    }

    public async resetASRService(): Promise<string> {
        return this.ssh.exec(`sudo sed -i "s#.*ExecStart=/usr/bin/${SnipsServices.asr}.*#ExecStart=/usr/bin/${SnipsServices.asr}#g" /lib/systemd/system/${SnipsServices.asr}.service; \
        sudo systemctl daemon-reload; \
        sudo systemctl restart ${SnipsServices.asr}.service`)
    }

    public async publishMQTTStartListening(): Promise<string> {
        return this.ssh.exec(`mosquitto_pub -h 127.0.0.1 -t 'hermes/asr/startListening' -m '{ "siteId":"default", "sessionId": "some_session" }'`)
    }

    public async killNCServer(): Promise<string> {
        return this.ssh.exec('pkill nc')
    }

    /*----------------------------------------------------------------------*/
    // RSA Key Generation

    public async sshCopyId(): Promise<string> {
        let sshPubKeyFile: string
        if (fsExtra.existsSync(snipsRSAKeyFilePath) === false) {
            sshPubKeyFile = await this.sshKeygen()
            this.addSnipsRSAKeyToSSHConfigIfNeeded()
        } else {
            sshPubKeyFile = fsExtra.readFileSync(snipsRSAKeyFilePath + '.pub', 'utf8')
            this.addSnipsRSAKeyToSSHConfigIfNeeded()
        }

        // The ssh-copy-id script doesn't exist on Windows, the Unix script essentially copies the keys on the remote server with ssh
        // You can check the original code on your own system : cat /usr/bin/ssh-copy-id
        // The command below is the same as the osx version.
        const command = `echo "${sshPubKeyFile}" | exec sh -c 'cd ; umask 077 ; mkdir -p .ssh && cat >> ${snipsRemoteAuthorizedKeysFilename} || exit 1 ; if type restorecon >/dev/null 2>&1 ; then restorecon -F .ssh ${snipsRemoteAuthorizedKeysFilename} ; fi'`
        return this.ssh.exec(command)
    }

    private ssh_keygen_binPath(): string {
        if (process.platform !== 'win32') { return 'ssh-keygen' }

        switch (process.arch) {
            case 'ia32': return path.join(__dirname, '..', 'bin', 'ssh-keygen-32.exe')
            case 'x64': return path.join(__dirname, '..', 'bin', 'ssh-keygen-64.exe')
        }

        throw new Error('Unsupported platform')
    }

    private async sshKeygen(): Promise<string> {
        const command = `${this.ssh_keygen_binPath()} -t rsa -b 2048 -C "Snips RSA key" -N "" -f ${snipsRSAKeyFilePath}`
        shelljs.exec(command, { async: false, silent: false })
        try {
            return fsExtra.readFileSync(snipsRSAKeyFilePath + '.pub', 'utf8')
        } catch (e) {
            throw new Error(`Snips RSA key couldn't be found`)
        }
    }

    private addSnipsRSAKeyToSSHConfigIfNeeded() {
        const sshConfigPath = os.homedir() + '/.ssh/config'
        const identityFilePath = '  IdentityFile ~/.ssh/' + snipsRSAKeyName

        let sshConfigFile: string | undefined
        try {
            const sshConfigFileLines = fsExtra.readFileSync(sshConfigPath, 'utf8').split('\n')
            if (sshConfigFileLines.indexOf(identityFilePath) >= 0) {
                return
            }
            const hostStarIndex = sshConfigFileLines.indexOf('Host *')
            if (hostStarIndex >= 0) {
                sshConfigFileLines.splice(hostStarIndex + 1, 0, identityFilePath)
            } else {
                sshConfigFileLines.push('Host *')
                sshConfigFileLines.push(identityFilePath)
            }
            sshConfigFile = sshConfigFileLines.join('\n')
        } catch (e) {
            sshConfigFile = `Host *\n  AddKeysToAgent yes\n${identityFilePath}`
        }
        fsExtra.writeFileSync(sshConfigPath, sshConfigFile)
    }
}
